/***************************************************************************
 *  Utility functions C library
 *  akiralib.c
 *
 *  by Akira Tsukamoto
 ***************************************************************************/

#include "sysdep.h"
#include "akiralib.h"

/***************************************************************************
 *** str2l() ***
 ** extended version of ANSI strtol() **
 ** ANSI strtol() will return 0 both on fail and string containing 0 value
 ** but str2l() will return -1 on fail
 * -> const char *string  : string to convert
 *    int radix           : radix, see ANSI C or POSIX strtol() for details
 *    long min            : limit of minimum value
 *    long max            : limit of maximum value
 * <- long *longInt       : result value
 *    int str2l()         : OK: 0  error: -1
 ***************************************************************************/
int str2l(long *result, const char *string, int radix, long min, long max)
{
  long lInt;
  char *endptr;

  if (result == NULL)
    return -1;
  if (string == NULL)
    return -1;
  if (radix < 0 || 36 < radix || radix == 1)
    return -1;
  if (min > max)
    return -1;

  lInt = strtol(string, &endptr, radix);

  /* Since ANSI C strtol() returns 0 for error
     chack whether string contains 0 for the value. */
  /*  Check if the sting was 0 */
  if (lInt == 0) {
    if (string == endptr)
      return -1;                /* genuine error */
    if (*endptr != '\0')
      return -1;                /* Had invalid character? */
  }
  if (lInt < min || max < lInt)
    return -1;

  *result = lInt;
  return 0;
}

/***************************************************************************
 *** openUDPPort() ***
 ** open UDP port with specified port number and returns fd **
 * -> in_port_t port           : port number
 * <- int openUDPPort()        : OK: fd  error: -1
 ***************************************************************************/
int openUDPPort(in_port_t port)
{
  int sockfd;
  struct sockaddr_in sockAddr;
  const int reuse = 1;          /* SO_REUSEADDR */

  /* open socket */
  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("openUDPPort(): socket()");
    return -1;
  }

  /* claer and set struct socket address */
  memset((char *) &sockAddr, 0, sizeof(sockAddr));
  sockAddr.sin_family = AF_INET;
  sockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  sockAddr.sin_port = port;

  /* bind */
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (const void *)&reuse,
             sizeof(reuse)) < 0) {
    perror("openUDPPort(): setsockopt(): SO_REUSEADDR");
    return -1;
  }
#ifdef SO_REUSEPORT
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEPORT, (const void *)&reuse,
             sizeof(reuse)) < 0) {
    perror("openUDPPort(): setsockopt(): SO_REUSEPORT");
    return -1;
  }
#endif
  if (bind(sockfd, (struct sockaddr *) &sockAddr, sizeof(sockAddr)) < 0) {
    perror("openUDPPort(): bind()");
    return -1;
  }

  return sockfd;
}

/** send UDP packets.
 * This is a wrapper function for sending UDP packets.
 * since UDP is not a relaible transfer protocol
 * this function will ignore error while sending and retruns no error */
void sendUDP(int sockfd, const void *data, size_t datasize,
            int flags, const struct sockaddr *to, socklen_t tolen)
{
  char *pdata = (char *)data;
  int written, retry = UDP_RETRY;

RESEND:
  written = sendto(sockfd, pdata, datasize, flags, to, tolen);
  /* if output buffer is full, retry */
  if (written == SOCKET_ERROR && ENOBUFS == sockerrno() && retry) {
    retry--;
    goto RESEND;
  }
  /* if all the data was not sent, resend only left data
   * but this should not happen in UDP */
  if ((size_t)written <  datasize && retry) {
    pdata    += written;
    datasize -= written;
    retry--;
    goto RESEND;
  }
}

/** WSAERRNO to errno */
int wsaetoe(int wsaerrno)
{
  #ifdef WIN32
  struct wse {
    int wsaerrno;
    int errno;
  };
  struct wse ws_errs[] = {
    {WSAEINTR          ,EINTR},
    {WSAEBADF          ,EBADF},
    {WSAEACCES         ,EACCES},
    {WSAEFAULT         ,EFAULT},
    {WSAEINVAL         ,EINVAL},
    {WSAEMFILE         ,EMFILE},
    {WSAEWOULDBLOCK    ,EWOULDBLOCK},
    {WSAEINPROGRESS    ,EINPROGRESS},
    {WSAEALREADY       ,EALREADY},
    {WSAENOTSOCK       ,ENOTSOCK},
    {WSAEDESTADDRREQ   ,EDESTADDRREQ},
    {WSAEMSGSIZE       ,EMSGSIZE},
//    {WSAEPROTOTYPE     ,EPROTOTYPE},
//    {WSAENOPROTOOPT    ,ENOPROTOOPT},
//    {WSAEPROTONOSUPPORT,EPROTONOSUPPORT},
//    {WSAESOCKTNOSUPPORT,ESOCKTNOSUPPORT},
//    {WSAEOPNOTSUPP     ,EOPNOTSUPP},
//    {WSAEPFNOSUPPORT   ,EPFNOSUPPORT},
//    {WSAEAFNOSUPPORT   ,EAFNOSUPPORT},
//    {WSAEADDRINUSE     ,EADDRINUSE},
    {WSAEADDRNOTAVAIL  ,EADDRNOTAVAIL},
//    {WSAENETDOWN       ,ENETDOWN},
//    {WSAENETUNREACH    ,ENETUNREACH},
//    {WSAENETRESET      ,ENETRESET},
//    {WSAECONNABORTED   ,ECONNABORTED},
//    {WSAECONNRESET     ,ECONNRESET},
    {WSAENOBUFS        ,ENOBUFS},
//    {WSAEISCONN        ,EISCONN},
//    {WSAENOTCONN       ,ENOTCONN},
//    {WSAESHUTDOWN      ,ESHUTDOWN},
//    {WSAETOOMANYREFS   ,ETOOMANYREFS},
//    {WSAETIMEDOUT      ,ETIMEDOUT},
//    {WSAECONNREFUSED   ,ECONNREFUSED},
//    {WSAELOOP          ,ELOOP},
    {WSAENAMETOOLONG   ,ENAMETOOLONG},
//    {WSAEHOSTDOWN      ,EHOSTDOWN},
//    {WSAEHOSTUNREACH   ,EHOSTUNREACH},
    {WSAENOTEMPTY      ,ENOTEMPTY},
//    {WSAEPROCLIM       ,EPROCLIM},
//    {WSAEUSERS         ,EUSERS},
//    {WSAEDQUOT         ,EDQUOT},
//    {WSAESTALE         ,ESTALE},
//    {WSAEREMOTE        ,EREMOTE},
    /* termination code */
    {0,0}
  };
  int i = 0;
  while(ws_errs[i].wsaerrno && ws_errs[i].wsaerrno != wsaerrno) {
    i++;
  }
  return ws_errs[i].errno;
  #else /* dummey return */
  return wsaerrno;
  #endif
}

/** socket errno.
 * wrapper function for returning socket specific errno,
 * this function was made because win32's error returning
 * is different from unix   */
int sockerrno(void)
{
  #ifndef WIN32
  return errno;
  #else
  return wsaetoe(WSAGetLastError());
  #endif
}

/***************************************************************************
 *** setSockAddr() ***
 ** set internet sockaddr(struct sockaddr_in) from IP and port number **
 * -> struct in_addr *ip        : pointer to IP number(struct in_addr)
 *    in_port_t port            : port number
 * <- struct sockaddr_in *sAddr : pointer to sockAddr,  No error return
 ***************************************************************************/
void setSockAddr(struct sockaddr_in *sAddr, struct in_addr *ip, in_port_t port)
{
  memset((char *) sAddr, 0, sizeof(struct sockaddr_in));
  sAddr->sin_family = AF_INET;
  sAddr->sin_addr.s_addr = ip->s_addr;
  sAddr->sin_port = port;
}

/***************************************************************************
 *** getLHnameIp() ***
 ** get Localhost Host name and IP address **
 * -> char *hName             : pointer to string buff
 *    int len                 : size of sting buff
 * <- char *hName             : local host name terminated by '\0'
 *    in_addr_t getLHnameIp() : OK: IP address   error: -1
 ***************************************************************************/
in_addr_t getLHnameIp(char *hName, size_t len)
{

  /* impliment code when host name was not retrieved */
  struct hostent *hostEnt;
  struct utsname myname;
  struct in_addr *ipAddr;

  if (uname(&myname) < 0)
    return -1;

  if ((hostEnt = gethostbyname(myname.nodename)) == NULL)
    return -1;
  ipAddr = (struct in_addr *) hostEnt->h_addr_list[0];
  if (strlen(myname.nodename) >= len)
    return -1;
  strcpy(hName, myname.nodename);

  return ipAddr->s_addr;
}

/***************************************************************************
 *** addr2str() ***
 ** convert sockaddr(struct sockaddr_in) to string
 ** if host name was retrieved
 **   sockaddr -> "www.yahoo.com(123.111.112.333)/4123"
 ** if not
 **   sockaddr -> "123.111.112.333/4123"
 * -> struct sockaddr_in *s           : sockaddr
 * <- char *addr2str(), char* string  : string,  No error return
 ***************************************************************************/
char *addr2str(char *string, size_t len, struct sockaddr_in *s)
{
  struct hostent *hostEnt;
  char temp[1024];

  hostEnt = gethostbyaddr((const char *) &s->sin_addr, sizeof(struct in_addr),
                          AF_INET);
  if (hostEnt != NULL) {
    if (ntohs(s->sin_port) != 0) {
      sprintf(temp, "%s(%s)/%d",
              hostEnt->h_name, inet_ntoa(s->sin_addr), ntohs(s->sin_port));
    } else {
      sprintf(temp, "%s(%s)/*****",
              hostEnt->h_name, inet_ntoa(s->sin_addr));
    }
  } else {
    if (ntohs(s->sin_port) != 0) {
      sprintf(temp, "%s/%d", inet_ntoa(s->sin_addr), ntohs(s->sin_port));
    } else {
      sprintf(temp, "%s/*****", inet_ntoa(s->sin_addr));
    }
  }

  if (strlen(temp) >= len) {
    return NULL;
  }

  strcpy(string, temp);
  return string;
}

/** check host.
 * check whether the sockAddr_in matches with ip and port
 * if port is 0, ignore port, check only ip
 * sockadd_in == (ip->s_addr, port)
 * match returns 0, false returns -1 */
int checkHost(struct sockaddr_in *sockAddr, struct in_addr *ip,
              in_port_t port)
{
  if (sockAddr->sin_addr.s_addr != ip->s_addr)
    return -1;
  if (port != 0) {
    if (sockAddr->sin_port != port)
      return -1;
  }

  return 0;
} /* checkHost() */

/** get only filename from whole path */
void getFilenameFromPath(char *filename, size_t len, char *path)
{
  char *p;

#ifndef WIN32
  if ( (p = strrchr(path, '/')) != NULL) {
     path = ++p;
  }
#else
  if ( (p = strrchr(path, '\\')) != NULL) {
     path = ++p;
  }
#endif
  strncpy(filename, path, NAME_MAX - 1);
  filename[NAME_MAX - 1] = '\0';
}

/** sleep milli sec. */
void sleepMsec(unsigned long msec)
{
  #ifndef WIN32  /* unix */
  struct timeval wait;
  wait.tv_sec  = (msec * 1000) / 1000000;
  wait.tv_usec = (msec * 1000) % 1000000;
  select(0, NULL, NULL, NULL, &wait);

  #else /* windows */
  Sleep(msec);
  #endif
}

/** convert 64bit ntp format to timeval format */
struct timeval ntp2tv(struct timeval *ret, uint32_t ntp_secs, uint32_t ntp_frac)
{
  /* I am ignoreing the high bit */
//  ret->tv_sec = ntp_secs & 0x7fffffff;
  ret->tv_sec = (ntp_secs - SECS_1900_1970_32) & 0x7fffffff;

  /* It will use only 20 bits of ntp_frac because it is enough for
   * 0 ... 999999 precision */
  ntp_frac >>= 12;

  /*
   *            ntp_frac >> 12 (20 bits)
   * tv_usec = ------------------------- * 1000,000 (-> nsec)
   *            0x10:0000      (20 bits)
   */

  /* if ntp_frac is small enough, calculate without using float */
  if (ntp_frac <= 274877) {
    ret->tv_usec = (ntp_frac * 15625) / 0x4000;
    return *ret;
  }
  ret->tv_usec = (long)((float)ntp_frac / 16384.0 * 15625.0);
  return *ret;
}

/** convert 32bit rtp time to timeval format */
struct timeval rtptime2tv(struct timeval *ret, uint32_t rtptime)
{
  ret->tv_sec  = (long)(rtptime >> 16);
  /* To eliminate using float, this uses the factorization
   * 10^6 / 2^16 == 15625 / 2^10 */
  ret->tv_usec = (long)(((rtptime & 0x0000FFFF) * 15625) >> 10);
  return *ret;
}

/** convert timeval format to 32bit rtp time */
uint32_t tv2rtptime(struct timeval *tv)
{
  uint32_t rtptime;

  rtptime = ((uint32_t)tv->tv_sec + SECS_1900_1970_32) << 16;
  /* To eliminate using float, this uses the factorization
   * 2^16 / 10^6 == 2^10 / 15625 */
  rtptime |= (uint32_t)((tv->tv_usec << 10) / 15625);
  return rtptime;
}

/*** addTimeval() ***/
struct timeval addTimeval(struct timeval *ret, struct timeval *a,
                          struct timeval *b)
{
  ret->tv_usec = a->tv_usec + b->tv_usec;
  if (ret->tv_usec >= 1000000) {
    ret->tv_sec = a->tv_sec + b->tv_sec + 1;
    ret->tv_usec -= 1000000;
  } else {
    ret->tv_sec = a->tv_sec + b->tv_sec;
  }
  return *ret;
}                               /* addTimeval() */

/*** subTimeval() ***/
struct timeval subTimeval(struct timeval *ret, struct timeval *a,
                          struct timeval *b)
{
  if (a->tv_usec < b->tv_usec) {
    ret->tv_usec = (a->tv_usec + 1000000) - b->tv_usec;
    ret->tv_sec = a->tv_sec - b->tv_sec - 1;
  } else {
    ret->tv_usec = a->tv_usec - b->tv_usec;
    ret->tv_sec = a->tv_sec - b->tv_sec;
  }
  return *ret;
}                               /* subTimeval() */

/*** maxTimeval() ***/
struct timeval maxTimeval(struct timeval *ret, struct timeval *a,
                           struct timeval *b)
{
  if (a->tv_sec > b->tv_sec) {
    *ret = *a;
  } else if (a->tv_sec < b->tv_sec) {
    *ret = *b;
  } else {                      /* a->tv_sec == b->tv_sec */
    if (a->tv_usec > b->tv_usec) {
      *ret = *a;
    } else {
      *ret = *b;
    }
  }
  return *ret;
}                               /* maxTimeval() */

/*** tv2double() ***/
double tv2double(struct timeval tv)
{
  double d;

  d = tv.tv_usec * 0.000001;
  d += tv.tv_sec;
  return d;
}                               /* tv2double() */

