/***************************************************************************
 *
 *  RTP Quality of Service measurement
 *  rtpqos.c
 *
 *  Copyright (c) 2002 by Columbia University; all rights reserved
 *  by Akira Tsukamoto
 *
 ***************************************************************************/

#include "sysdep.h"
#include "notify.h"
#include "multimer.h"
#include "host2ip.h"
#include "queue.h"
#include "akiralib.h"
#include "rtpqos.h"
#include "rtp_highlevel.h"

/*** global Variables ******************************************************/

/** getopt() */
extern char *optarg;
extern int optind;

/*** for debugging    ******************************************************/

//#define DEBUG_NO_SEND 1
static int debug;

/*** private Variables *****************************************************/

/** flag for flushing packets in the queue
 * flush = 1: do not insert any more packets in the queue */
static int flush;

/** holds all the parameters specified from command line */
static Options g_o;

/** holds the RTP descriptor */
static Rtp g_rtp;

/** holds all the statistics */
static Statistics g_stat;

/*
 * Payload type map. from rtpdump.c
 */
static struct {
  char *enc;      /* encoding name */
  int  rate;      /* sampling rate for audio; clock rate for video */
  int  ch;        /* audio channels; 0 for video */
} pt_map[256];

/*** prototypes for private functions used in main() *************************/
static Notify_value sendRTCPHandler(Notify_client client);

/***************************************************************************
 *** Private function definition ***
 ***************************************************************************/

/*
 * Set up payload type map. We should be able to read this in
 * from a file.
 * This function was extracted from the code of rtpdump.c */
static void setupRTPPtTable(void)
{
  int i;

  for (i = 0; i < 256; i++) {
    pt_map[i].enc  = "????";
    pt_map[i].rate = 0;
    pt_map[i].ch   = 0;
  }
  /* Updated 11 May 2002 by Akira Tsukamoto with current IANA assignments: */
  /* http://www.iana.org/assignments/rtp-parameters */
  /* Marked *r* items are indicated as 'reserved' by the IANA */
  pt_map[  0].enc = "PCMU"; pt_map[  0].rate =  8000; pt_map[  0].ch = 1;
  pt_map[  1].enc = "1016"; pt_map[  1].rate =  8000; pt_map[  1].ch = 1;
  pt_map[  2].enc = "G721"; pt_map[  2].rate =  8000; pt_map[  2].ch = 1;
  pt_map[  3].enc = "GSM "; pt_map[  3].rate =  8000; pt_map[  3].ch = 1;
  pt_map[  4].enc = "G723"; pt_map[  4].rate =  8000; pt_map[  4].ch = 1;
  pt_map[  5].enc = "DVI4"; pt_map[  5].rate =  8000; pt_map[  5].ch = 1;
  pt_map[  6].enc = "DVI4"; pt_map[  6].rate = 16000; pt_map[  6].ch = 1;
  pt_map[  7].enc = "LPC "; pt_map[  7].rate =  8000; pt_map[  7].ch = 1;
  pt_map[  8].enc = "PCMA"; pt_map[  8].rate =  8000; pt_map[  8].ch = 1;
  pt_map[  9].enc = "G722"; pt_map[  9].rate =  8000; pt_map[  9].ch = 1;
  pt_map[ 10].enc = "L16 "; pt_map[ 10].rate = 44100; pt_map[ 10].ch = 2;
  pt_map[ 11].enc = "L16 "; pt_map[ 11].rate = 44100; pt_map[ 11].ch = 1;
  pt_map[ 12].enc = "QCELP"; pt_map[ 12].rate = 8000; pt_map[ 12].ch = 1;
  pt_map[ 14].enc = "MPA "; pt_map[ 14].rate = 90000; pt_map[ 14].ch = 0;
  pt_map[ 15].enc = "G728"; pt_map[ 15].rate =  8000; pt_map[ 15].ch = 1;
  pt_map[ 16].enc = "DVI4"; pt_map[ 16].rate = 11025; pt_map[ 16].ch = 1;
  pt_map[ 17].enc = "DVI4"; pt_map[ 17].rate = 22050; pt_map[ 17].ch = 1;
  pt_map[ 18].enc = "G729"; pt_map[ 18].rate =  8000; pt_map[ 18].ch = 1;
  pt_map[ 23].enc = "SCR "; pt_map[ 23].rate = 90000; pt_map[ 23].ch = 0; /*r*/
  pt_map[ 24].enc = "MPEG"; pt_map[ 24].rate = 90000; pt_map[ 24].ch = 0; /*r*/
  pt_map[ 25].enc = "CelB"; pt_map[ 25].rate = 90000; pt_map[ 25].ch = 0;
  pt_map[ 26].enc = "JPEG"; pt_map[ 26].rate = 90000; pt_map[ 26].ch = 0;
  pt_map[ 27].enc = "CUSM"; pt_map[ 27].rate = 90000; pt_map[ 27].ch = 0; /*r*/
  pt_map[ 28].enc = "nv  "; pt_map[ 28].rate = 90000; pt_map[ 28].ch = 0;
  pt_map[ 29].enc = "PicW"; pt_map[ 29].rate = 90000; pt_map[ 29].ch = 0; /*r*/
  pt_map[ 30].enc = "CPV "; pt_map[ 30].rate = 90000; pt_map[ 30].ch = 0; /*r*/
  pt_map[ 31].enc = "H261"; pt_map[ 31].rate = 90000; pt_map[ 31].ch = 0;
  pt_map[ 32].enc = "MPV "; pt_map[ 32].rate = 90000; pt_map[ 32].ch = 0;
  pt_map[ 33].enc = "MP2T"; pt_map[ 33].rate = 90000; pt_map[ 33].ch = 0;
  pt_map[ 34].enc = "H263"; pt_map[ 34].rate = 90000; pt_map[ 34].ch = 0;
  
} /* setupRTPPtTable() */

/** initialize options.
 * set the default value to struct Options */
static int initOptions(Options *o, char *argv0)
{
  struct hostent *host;
  struct in_addr *local;

  /* get local IP address for default dest host */
  if ((host = gethostbyname("localhost")) == NULL) {
    perror("initOptions(): gethostbyname()");
    return -1;
  }
  local = (struct in_addr *) host->h_addr_list[0];

  /* create RTP payload table */
  setupRTPPtTable();

  getFilenameFromPath(o->argv0, NAME_MAX, argv0);
  strncpy(o->logname, DEFAULT_LOG, NAME_MAX - 1);
  o->logname[NAME_MAX - 1] = '\0';
  o->payloadType  = DEFAULT_PAYLOAD_TYPE;
  o->samplingRate = pt_map[g_o.payloadType].rate;
  o->packInterval = DEFAULT_PZT;
  o->destHostIpN  = *local;
  o->verbose      = 0;

  return 0;
} /* initOptions() */

/** initialize statistics.
 * set the default value to struct Statistics */
static int initStat(Statistics *s)
{
  struct timeval now_tv;

  gettimeofday(&now_tv, NULL);
  s->rtp_startime = s->rtp_lasttime = now_tv;
  s->rtcp_startime = s->rtcp_lasttime = now_tv;

  if ((s->queue = openQueue()) == NULL) {
    fprintf(stderr, "%s: Could not create Queue.\n", g_o.argv0);
    return -1;
  }

  return 0;
} /* initStat() */

/** close statistics.
 * destroy struct stat */
static int closeStat(Statistics *stat)
{
  closeQueue(stat->queue);

  return 0;
} /* closeStat() */

/***************************************************************************/

/** get host name and port number.
 * parse the commandline string and return hostIpN and hostPorts */
static int getHost(struct in_addr *hostIpN,
                   in_port_t * hostPort_a, in_port_t * hostPort_b, char *optarg)
{
  char *pStr, buff[STRSIZE];
  struct in_addr ip;
  long port;       /* since in_port_t is 16 bit, it is large enough */

  if (strlen(optarg) >= STRSIZE) {
    fprintf(stderr, "String is too large: %s\n", optarg);
    return -1;
  }
  strcpy(buff, optarg);

  /* first token (hostIpN) */
  pStr = strtok(buff, ":/");
  if (pStr != NULL) {
    ip = host2ip(pStr);
    if (ip.s_addr == INADDR_ANY)
      return -1;
    *hostIpN = ip;
  }

  /* second token (hostPort_a) */
  pStr = strtok(NULL, ":/");
  if (pStr != NULL) {
    if (str2l(&port, pStr, 0, 0, 0xFFFF) == -1)
      return -1;
    *hostPort_a = port;
  }

  /* third token (hostPort_b) */
  pStr = strtok(NULL, ":/");
  if (pStr != NULL) {
    if (str2l(&port, pStr, 0, 0, 0xFFFF) == -1)
      return -1;
    *hostPort_b = port;
  }

  if (debug)
    printf("option: IP and Port: %s(0x%lX)/%d/%d\n",
           inet_ntoa(*hostIpN), (lu)hostIpN->s_addr, *hostPort_a, *hostPort_b);
  return 0;
} /* getHost() */

/** usage.
 * Just prints out the usage */
static void usage(char *argv0)
{
  fprintf(stdout, "\
Usage: %s [-v] [-h] [-o file_name] [-i packetizing_interval] \n\
          remote_host/local_port[/remote_port]\n",
  argv0);
} /* usage() */

/** parse command line.
 * this is the main getopt() routine */
static int parseCmdLine(int argc, char *argv[])
{
  int c, errF = 0;
  long l;

  while ( (c = getopt(argc, argv, "o:i:m:p:l:nbvhd")) != EOF) {
    switch (c) {
    case 'o':
      strncpy(g_o.logname, optarg, NAME_MAX - 1);
      g_o.logname[NAME_MAX - 1] = '\0';
      break;
    case 'i':
      if (str2l(&l, optarg, 0, 1, MAX_PACK_INTERVAL) == -1)
        errF = -1;
      g_o.packInterval = (int)l;
      break;
    case 'm':
      if (str2l(&l, optarg, 0, 0, 1) == -1)
        errF = -1;
      g_o.qos_mode = (int)l;
      break;
    case 'p':
      if (str2l(&l, optarg, 0, 0, MAX_PAYLOAD_TYPE) == -1)
        errF = -1;
      g_o.payloadType = (int)l;
      break;
    case 'l':
      if (str2l(&l, optarg, 0, 1, MAX_PROVE_SEC) == -1)
        errF = -1;
      g_o.probe_sec = (int)l;
      break;
    case 'n':
      g_o.non_interactive++;
      break;
    case 'b':
      g_o.loopback_ok++;
      break;
    case 'v':
      g_o.verbose++;
      break;
    case '?':
    case 'h':
      usage(g_o.argv0);
      exit(1);
      break;
    case 'd':                   /* for debugging */
      debug++;
      break;
    } /* switch */

    if (errF == -1) {
      fprintf(stderr, "%s: Invalid option: -%c %s\n\n", g_o.argv0, c, optarg);
      usage(g_o.argv0);
      exit(1);
    }
  } /* while */

  if (argc - optind != 1) {
    fprintf(stderr, "%s: remote_host/local_port is missing.\n", g_o.argv0);
    usage(g_o.argv0);
    exit(1);
  }
  /* parse "host_name/port" */
  if (getHost(&g_o.destHostIpN, &g_o.localPort, &g_o.destHostPort, argv[argc-1])
      == -1) {
    fprintf(stderr, "%s: Invalid option: %s\n", g_o.argv0, argv[argc-1]);
    usage(g_o.argv0);
    exit(1);
  }
  strncpy(g_o.destHost, argv[argc-1], NAME_MAX - 1);

  return 0;
} /* parseCmdLine() */

/***************************************************************************/

/** open log file.
 * opens two log file with ".rtp" and ".rtcp" postfix
 * returns 0 on success and -1 on fail */
static int openLog(char *logFile)
{
  char logFileName[NAME_MAX];

  strcpy(logFileName, logFile);
  strcat(logFileName, ".rtp");
  if ( (g_rtp.RTPlog = fopen(logFileName, "wb")) == NULL) {
    perror(logFileName);
    return -1;
  }

  strcpy(logFileName, logFile);
  strcat(logFileName, ".rtcp");
  if ( (g_rtp.RTCPlog = fopen(logFileName, "wb")) == NULL) {
    perror(logFileName);
    return -1;
  }

  return 0;
} /* openLog() */

/* This function was referred from the example1/2.c from
 * RTP Library Copyright 1997, 1998 Lucent Technologies */
/* Set up an RTP Library connection.
 * Return the maximum file descriptor number used, or -1 on error.
 * remote should contain the address to connect to in
 * host/port[/ttl] format.  (If ttl is not specified, '1' is used.)
 * cid will contain the context ID of the RTP session on successful return.
 * sock will contain the file descriptors of the RTP and RTCP ports.
 *
 * BUG: on failure, we should tear down the partially-completed
 * connections. */
static int setup_connection(
  context *cid, int *rtpFd, int *rtcpFd,
  char *remote_host, u_int16 remote_port, u_int16 local_port)
{
  /* Error values that the RTP library returns. */
  rtperror err;
  /* The time-to-live value to set if we're talking to a multicast
   * address. */
  unsigned char ttl = 1;
  /* The sockaddr that hpt returns to us.  Note that we don't actually
   * pass this value to the library; it's just a convenient structure. */
  /* The type the library uses to keep track of sockets */
  socktype sockt; 
  /* The highes file descriptor we've dealt with */
  int nfds = 0;
  
  /* Some values for creating the CNAME (see below). */
  char *username;

  char *sp;

  char hostname[MAXHOSTNAMELEN];
  char cname[MAXHOSTNAMELEN+32];
  char name[MAXHOSTNAMELEN+32];
  struct timeval tp;
  person uid, conid;

  /* initialize parameters */
  sp = strchr(remote_host, '/');
  if (sp != NULL)
    *sp = '\0';
  ttl  = 1;

  /* Create the RTP session -- allocate data structures.
   * This doesn't open any sockets. */
  err = RTPCreate(cid);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }

  /* Set destination host */

  /* Set the RTP Session's host address to send to. */
  err = RTPSessionAddSendAddr(*cid, remote_host, remote_port, ttl);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }

  /* Set the RTP Session's local receive address */
  /* We set this to the dest for the following reason:
     - It's necessary if the dest is a multicast group, so we join the group
     - It's okay if the dest is unicast, since setting a dest which isn't a
       local address is ignored. */
  err = RTPSessionSetReceiveAddr(*cid, remote_host, local_port);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }

  /* Set user info */

  /* Set up our CNAME (canonical name): should be of the form user@host */
  username = getenv("USER");
  if (gethostname(hostname, MAXHOSTNAMELEN) < 0) {
    perror("gethostname");
    return -1;
  }
  if (username) { sprintf(cname, "%s@%s", username, hostname); }
  else { strcpy(cname, hostname); }

  /* Member 0 is the local member (us) */
  err = RTPMemberInfoSetSDES(*cid, 0, RTP_MI_CNAME, cname);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }

  /* Set up our NAME (standard display name) */
  if (username) { sprintf(name, "%s-%s", DEFAULT_RTP_NAME, hostname); }
  else { strcpy(name, DEFAULT_RTP_NAME); }
  err = RTPMemberInfoSetSDES(*cid, 0, RTP_MI_NAME, name);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }

  /* Add localhost as a single contributor */

  gettimeofday(&tp, NULL);
  srand(tp.tv_usec);
  conid = rand();
  err = RTPSessionAddToContributorList(*cid, (int) conid);
  if (err) {
    fprintf(stderr, "%s %s\n", RTPStrError(err), RTPDebugStr());
    return -1;
  }
  err = RTPSessionGetUniqueIDForCSRC(*cid,conid,&uid);
  if (err) {
    fprintf(stderr, "%s %s\n", RTPStrError(err), RTPDebugStr());
    return -1;
  }
  sprintf(cname, "Contributor %d", (int) conid);
  err = RTPMemberInfoSetSDES(*cid, uid, RTP_MI_CNAME, cname);
  if (err) {
    fprintf(stderr, "%s %s\n", RTPStrError(err), RTPDebugStr());
    return -1;
  }
  err = RTPMemberInfoSetSDES(*cid, uid, RTP_MI_NAME, cname);
  if (err) {
    fprintf(stderr, "%s %s\n", RTPStrError(err), RTPDebugStr());
    return -1;
  }

  /* Establish a connection */

  /* Open the connection.  We're now live on the network. */
  err = RTPOpenConnection(*cid);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }

  /* Get the socket that RTP data is transmitted on, so we can select() on
   * it. */
  /* Note that the value passed to it is a socktype *, not an int *. */
  err = RTPSessionGetRTPSocket(*cid, &sockt);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }
  *rtpFd = sockt;
  nfds = 0;
#ifdef __unix
  if (nfds < sockt) nfds = sockt;
#endif

  /* Get the socket that RTCP control information is transmitted on, so we
     can select() on it. */
  err = RTPSessionGetRTCPSocket(*cid, &sockt);
  if (err != RTP_OK) {
    fprintf(stderr, "%s\n", RTPStrError(err));
    return -1;
  }
  *rtcpFd = sockt;

#ifdef __unix
  if (nfds < sockt) nfds = sockt;
#endif

  return nfds;
} /* setup_connection */

/** initialize RTP.
 * initialize RTP packet descriptor
 * returns 0 on success and -1 on fail */
static int initRtp(Rtp *r)
{
  struct timeval now_tv;

  if (setup_connection(&g_rtp.cid, &g_rtp.rtp_fd, &g_rtp.rtcp_fd,
                       g_o.destHost, g_o.destHostPort, g_o.localPort) < 0) {
    fprintf(stderr, "%s: initRtp(): setup_connection()\n", g_o.argv0);
    return -1;
  }

  /* calc packInterval_tv */
  r->packInterval_tv.tv_usec = g_o.packInterval * 1000;
  if (debug)
    printf("g_rtp.packInterval_tv: %f\n", tv2double(r->packInterval_tv));

  /* initial departure time */
  gettimeofday(&now_tv, NULL);
  timeradd(&now_tv, &r->packInterval_tv, &r->dept_tv);

  /* calc absolute ending time */
  g_o.stop_tv_sec = now_tv.tv_sec + g_o.probe_sec;

  /* calc RTP packet payload size */
  r->datasize = g_o.samplingRate * g_o.packInterval / 1000;
  if (debug)
    printf("g_o.samplingRate: %d g_o.packInterval: %d g_rtp.datasize: %d\n", g_o.samplingRate, g_o.packInterval, r->datasize);

  return 0;
} /* initRtp() */

/** close RTP
 * destroy RTP descriptor
 * returns 0 on success, -1 on fail */
static int closeRtp(Rtp *r)
{
  rtperror err;

  /* We're done sending now, so close the connection. */
  /* This sends the RTCP BYE packet and closes our sockets. */
  if ((err = RTPCloseConnection(g_rtp.cid,"Goodbye!")) != RTP_OK) {
    fprintf(stderr, "closeRtp(): RTPCloseConnection(): %s\n", RTPStrError(err));
    return -1;
  }

  /* De-allocate all the data structures reserved for the connection.
   * (This would also close the connection, as well, if we hadn't just done
   * it above.) */
  if ((err = RTPDestroy(g_rtp.cid)) != RTP_OK) {
    fprintf(stderr, "closeRtp(): RTPDestroy(): %s\n", RTPStrError(err));
    return -1;
  }

  return 0;
} /* closeRtp() */

/** initialize.
 * constructor for main().
 * this will prepare all the initial setup for main()
 * opening UDP port, creating queue end etc */
static int init(Options * options)
{
  /* open a debug logging file */
  if (debug >= 2) {
    if ( (g_o.debfile = fopen("debug.log", "wb")) == NULL) {
      perror("init(): fopen(debug.log)");
      return -1;
    }
  }

  /* open rtp handler */
  /* if the destination port is not specified then use the
   * same with the local binding port */
  if (debug)
    printf("g_o.destHostPort: %d g_o.localPort: %d\n",
                              g_o.destHostPort, g_o.localPort);
  if (g_o.destHostPort == 0)
    g_o.destHostPort = g_o.localPort;
    if (debug)
      printf("g_o.destHostPort: %d g_o.localPort: %d\n",
                                g_o.destHostPort, g_o.localPort);
  if (initRtp(&g_rtp) < 0) {
    fprintf(stderr, "%s: init(): initRtp().\n", g_o.argv0);
    return -1;
  }

  /* open log file */
  if (openLog(g_o.logname) < 0) {
    fprintf(stderr, "%s: init(): openLog().\n", g_o.argv0);
    return -1;
  }

  /* set default value to statistics struct */
  initStat(&g_stat);

  return 0;
} /* init() */

/** destroy.
 * destructor for main().
 * this will close all the resource was used in main()
 * deleting RTP handler, closing UDP port */
static int destroy(void)
{
  if (closeRtp(&g_rtp) < 0) {
    fprintf(stderr, "%s: destroy(): closeRtp().\n", g_o.argv0);
    return -1;
  }

  /* close log files */
  if (fclose(g_rtp.RTPlog) < 0)
    perror("closing RTP logfile");
  if (fclose(g_rtp.RTCPlog) < 0)
    perror("closing RTP logfile");

  closeStat(&g_stat);

  if (debug >= 2)
    if (fclose(g_o.debfile) < 0)
      perror("closing debug logfile");

  return 0;
} /* destroy() */

/***************************************************************************/
/** print title.
 * Just print out the title */
static int printTitle(Options * options)
{
  if (g_o.verbose == 0) goto NOPRINT;

  puts  ("RTP Quality of Service");
  puts  (" Copyright (c) 2002 by Columbia University; all rights reserved");
  putchar('\n');

  NOPRINT:
  return 0;
} /* printTitle() */

/** print statistics.
 * print out all the statistics after stopping this emulator */
static void printStatistics(void)
{
  if (g_o.verbose == 0) goto NOPRINT;

  putchar('\n');
  puts  ("--- Statistics ---");
  puts  ("RTP packets:");
  printf("  start time: %.6f\n", tv2double(g_stat.rtp_startime));
  printf("  last  time: %.6f\n", tv2double(g_stat.rtp_lasttime));
  printf("  total     : %d(packets)\n", g_stat.recvRTPPack);
  if (g_o.qos_mode != 0) goto NORTT;
  printf("  rtt  max  : %f(sec)\n", g_stat.rtp_rttMax / 1000000.0);
  printf("       avr  : %f(sec)\n", g_stat.rtp_rttAvr / 1000000.0);
  NORTT:
  printf("  diff max  : %f(%%)\n", g_stat.rtp_difMax * 100);
  printf("       avr  : %f(%%)\n", g_stat.rtp_difAvr * 100);
  printf("  jitter    : %d\n", g_stat.rtp_jitter);
  printf("  loss      : %d(packets)\n", g_stat.rtp_loss);
  putchar('\n');
  puts  ("RTCP packets:");
  printf("  start time: %.6f\n", tv2double(g_stat.rtcp_startime));
  printf("  last  time: %.6f\n", tv2double(g_stat.rtcp_lasttime));
  printf("  total     : %d(packets)\n", g_stat.recvRTCPPack);
  printf("  rtt  max  : %f(sec)\n", g_stat.rtcp_rttMax / 1000000.0);
  printf("       avr  : %f(sec)\n", g_stat.rtcp_rttAvr / 1000000.0);
  putchar('\n');
  NOPRINT:
  return;
} /* printStatistics() */

/***************************************************************************/

/** assemble item.
 * pack seq number and dep_tv into qItem and return it */
static QItems *assembleItem(uint32_t seq, struct timeval *dep_tv)
{
  QItems *qItem;

  qItem = (QItems *) malloc(sizeof(QItems));
  if (qItem == NULL) {
    perror("assembleItem(): malloc()");
    return NULL;
  }
  qItem->seq    = seq;
  qItem->dep_tv = *dep_tv;

  return qItem;
} /* assembleItem() */

/** disassemble item.
 * unpack seq number and dep_tv from qItem */
static void disAssembleItem(uint32_t *seq, struct timeval *dep_tv,
                            QItems * qItem)
{
  *seq    = qItem->seq;
  *dep_tv = qItem->dep_tv;

  free(qItem);
} /* disAssembleItem() */

/** coredump the packet */
static void coreDumpPacket(char *rBuff, int rBytes, int packet_type)
{
  int i;

  fprintf(g_o.debfile, "%s: total_nr=%d: data=",
                        packet_type ? "RTCP" : "RTP", g_stat.recvRTPPack);
  for (i = 0; i < rBytes; i++) {
    fprintf(g_o.debfile, "%02hx ", (unsigned short int)(unsigned char)rBuff[i]);
  }
  putc('\n', g_o.debfile);

} /* coreDumpPacket() */

/* This function will computes the interval time in milli-sec
 * from rtp-payload-type and timestamp */
static int getInterval(rtp_hdr_t *hdr)
{
  static int      first_packet = 1;
  uint32_t        ts;
  int samp_rate;
  int interval;
  int pt = (int)RTP_PT(*hdr);

  if (!(samp_rate = pt_map[pt].rate))
    samp_rate = DEFAULT_SAMPLING;

  if (first_packet) {
    g_stat.prev_ts = hdr->ts - g_o.packInterval * samp_rate / 1000;
    first_packet = 0;
  }

  ts = hdr->ts;

  /* timestamp * 1000
   * ---------------- = interval
   *   sampling rate
   */
  if (debug >= 2) printf("pt_map[%d].rate: %d ts_hex: %08lX ts: %lu\n",
                    pt, samp_rate, (lu)hdr->ts, (lu)(ts - g_stat.prev_ts));

  interval = (int)(ts - g_stat.prev_ts) * 1000 / samp_rate;
  return interval;
}

/* extend 8bit ID to 32bit seq.
 * every time 8bit ID overlaps, the upper 24 bit of seq will be
 * incremented */
static uint32_t id2seq(uint8_t id)
{
  static uint32_t upper24;
  uint32_t seq;
  static uint8_t lastid;
  uint8_t gap;

  gap = id - lastid;
  if((gap & 0xff) < MAX_RTP_DROPOUT && id < lastid)
    upper24++;

  lastid = id;
  seq = upper24 << 8 | (id & 0xff);

  if (debug >= 3)
    printf("id2seq(): seq: %x id: %x\n", (unsigned int)seq, id);

  return seq;
} /* id2seq() */

/* convert 32bit seq to 8bit id */
static uint8_t seq2id(uint32_t seq)
{
  return (uint8_t)(seq & 0xFF);
} /* seq2id() */

/** get ID number from the packet.
 * returns ID for matching packets */
static uint32_t getIDFromPacket(char *buff)
{
  uint32_t id = 0;

  id = (uint32_t)(unsigned char)buff[12];
  if (debug >= 2)
    fprintf(g_o.debfile, "ID=%lu: ", (lu)id);

  return id;
} /* getIDFromPacket() */

/* is old packet
 * Check whether the seq number have reversed from out of order 
 * Should be safe for overlapping 32bit counter */
static int isOldPacket(uint32_t last_recvSeq, uint32_t recvSeq)
{
  uint32_t gap;

  gap = recvSeq - last_recvSeq;
  if (gap < MAX_RTP_DROPOUT)
    return 0;
  else
    return 1;     /* old packet */
} /* isOldPacket */

static void getUpdateDiffJitter(float *diff, struct timeval *interval_tv,
                                struct timeval *now_tv, rtp_hdr_t *hdr)
{
  /* initialize arrival time of the previous packet, only on frist receive */
  if ((g_stat.prev_arrive_tv.tv_sec & g_stat.prev_arrive_tv.tv_usec) == 0)
    timersub(now_tv, &g_rtp.packInterval_tv, &g_stat.prev_arrive_tv);

  /* calc actual interval.
   * arrival_time - prev_arrival_time
   */
  timersub(now_tv, &g_stat.prev_arrive_tv, interval_tv);

  /* calc difference 
   * |         actual_interval(usec)             |
   * | ------------------------------------- - 1 |
   * | ideal_interval(msec) * 1000 (to usec)     |
   */
  *diff = (float)(interval_tv->tv_sec * 1000000L + interval_tv->tv_usec)
          / (getInterval(hdr) * 1000.0) - 1.0;

  /* update diff */
  if (g_stat.rtp_difMax < *diff)
    g_stat.rtp_difMax = *diff;
  g_stat.rtp_difAvr = (g_stat.rtp_difAvr * g_stat.recvRTPPack + *diff)
                      / (g_stat.recvRTPPack + 1.0);

  /* update jitter.
   * new_jitter = old_jitter + (|diff| - old_jitter) / 16
   */
  g_stat.rtp_jitter
    = (int)(g_stat.rtp_jitter
            + ((*diff < 0? -(*diff) * 100: *diff * 100) - g_stat.rtp_jitter)
            / 16.0);
}

/** log interval, diff and jitter.*/
static void logDiffJitter(struct timeval *now_tv, rtp_hdr_t *hdr)
{
  static unsigned int id;
  struct timeval interval_tv;
  float diff = 0;

  getUpdateDiffJitter(&diff, &interval_tv, now_tv, hdr);

  /* log to the file */
  fprintf(g_rtp.RTPlog, "%lu %ld.%05ld %.5f %d\n",
                        (lu)id,
                        interval_tv.tv_sec, interval_tv.tv_usec / 10, diff,
                        g_stat.rtp_jitter);
  if (debug >= 2) {
    fprintf(g_o.debfile,
            "id=%lu interval=%ld.%06ld diff=%f |diff|=%f rtp_jitter=%d \n",
                        (lu)id,
                        interval_tv.tv_sec, interval_tv.tv_usec / 10,
                        diff, (diff < 0? -diff: diff),
                        g_stat.rtp_jitter);
  }

  id++;

} /* logDiffJitter */

/** log round trip time, interval, diff and jitter. */
static void logRttDiffJitter(uint32_t id, struct timeval *now_tv,
                             struct timeval *sent_tv, rtp_hdr_t *hdr)
{
  struct timeval rtt_tv, interval_tv;
  unsigned long  rtt;
  float diff = 0;

  /* calc rtt */
  timersub(now_tv, sent_tv, &rtt_tv);
  rtt = rtt_tv.tv_sec * 1000000 + rtt_tv.tv_usec;

  /* update rtt */
  if (g_stat.rtp_rttMax < rtt)
    g_stat.rtp_rttMax = rtt;
  g_stat.rtp_rttAvr = (g_stat.rtp_rttAvr * g_stat.recvRTPPack + rtt)
                      / (g_stat.recvRTPPack + 1);

  getUpdateDiffJitter(&diff, &interval_tv, now_tv, hdr);

  /* log to the file */
  fprintf(g_rtp.RTPlog, "%lu %ld.%05ld %ld.%05ld %.5f %d %d\n",
                        (lu)id, rtt_tv.tv_sec, rtt_tv.tv_usec / 10,
                        interval_tv.tv_sec, interval_tv.tv_usec / 10, diff,
                        g_stat.rtp_jitter, g_stat.rtp_loss);
  if (debug >= 2) {
    fprintf(g_o.debfile,
"id=%lu rtt=%ld.%06ld interval=%ld.%06ld diff=%f |diff|=%f rtp_jitter=%d \
rtp_loss=%d\n",
                        (lu)id, rtt_tv.tv_sec, rtt_tv.tv_usec,
                        interval_tv.tv_sec, interval_tv.tv_usec,
                        diff, (diff < 0? -diff: diff),
                        g_stat.rtp_jitter, g_stat.rtp_loss);
  }
} /* logRttDiffJitter */

/** match packet
 * match packet with sent packet in the queue
 * This code is unorganized and rough, need to be fixes some times */
static int matchPacket(uint32_t recvSeq, struct timeval *now_tv, rtp_hdr_t *hdr)
{
  static uint32_t last_recvSeq;
  QItems *qItem;
  uint32_t sentSeq;
  struct timeval dep_tv;
  int retry = 0;        /* this needs to be fixed */

  /* reject invalid recvSeq packet */
  if (debug)
    printf("recvseq - last_recvseq: %u %u %u\n",
           (unsigned int)recvSeq, (unsigned int)last_recvSeq, 
           (unsigned int)(recvSeq - last_recvSeq));
  if (recvSeq - last_recvSeq > MAX_RTP_DROPOUT)
    return 0;

  RETRY:
  /* get the sent time for this seq from the queue */
  if (getQueue(g_stat.queue, &qItem) == NULL) {
    fprintf(stderr, "%s: calcStatistics(): getQueue():\n", g_o.argv0);
    return -1;
  }
  disAssembleItem(&sentSeq, &dep_tv, qItem);

  if (debug >= 2) {
    fprintf(g_o.debfile, "matching with sent packet: ");
    fprintf(g_o.debfile, "sentSeq=%lu recvSeq=%lu ", (lu)sentSeq, (lu)recvSeq);
    fprintf(g_o.debfile, "dep_tv=%ld.%06ld now_tv=%ld.%06ld ",
            dep_tv.tv_sec, dep_tv.tv_usec, now_tv->tv_sec, now_tv->tv_usec);
  }

  if (sentSeq == recvSeq) {
    /* packet matched with id */

    logRttDiffJitter(recvSeq, now_tv, &dep_tv, hdr);

  } else if (isOldPacket(last_recvSeq, recvSeq)) {
    /* out of order occurred previously and the an old packet arrived
     * just discard this packet */

    if (g_o.verbose) {
      printf("received old packets from out of order: seq: %u\n",
             (unsigned int)recvSeq);
      if (debug)
        printf("  sentSeq: %lu recvSeq: %lu last_recvSeq %lu\n",
               (lu)sentSeq, (lu)recvSeq, (lu)last_recvSeq);
    }

    /* put the seq num back to the tail of the queue */
    qItem = assembleItem(sentSeq, &dep_tv);
    if (addTailQueue(g_stat.queue, qItem) == NULL) {
      fprintf(stderr, "%s: calcStatisitcs(): addTailQueue():\n", g_o.argv0);
      return -1;
    }

  } else {
    /* packet loss or out of order occurred and seq number was jumped */

    if (g_o.verbose) {
      printf("detected packet loss or out of order: %u(packets) jumped\n",
              (unsigned int)(recvSeq - last_recvSeq));
      if (debug)
        printf("  sentSeq: %lu recvSeq: %lu last_recvSeq %lu\n",
               (lu)sentSeq, (lu)recvSeq, (lu)last_recvSeq);
    }

    g_stat.rtp_loss++;

    /* get the next sentSeq to match received packet */
    if (countQueue(g_stat.queue) > 0 && retry < 10) {
      retry++;
      goto RETRY;
    }

    if (debug)
        puts("sentSeq is empty");
  }

  last_recvSeq = recvSeq;

  if (debug >= 2)
    putc('\n', g_o.debfile);

  return 0;
} /* matchPacket() */

/* QOS measurement from RTP packet
 * calculate QOS statistics */
static int qosRTP(rtp_packet *rtp, char *rBuff, int rBytes,
                  struct timeval *now_tv)
{
  static int first_received = 1;
  static uint8_t last_id;
  uint32_t id;

  switch (g_o.qos_mode) {
  case 0: /* match mode */
    /* retrieve seq num from the content of the packet */
    id = getIDFromPacket(rBuff);

    /* Some times, junk packets arrives before our first sent packet comes back.
     * Needs to discard it */
    if (first_received) {
      if (id != 0)
        return 0;          /* discard */
      else
        first_received = 0; /* our first sent packet arrived */
    }

    /* This is a hack, but sipconf returns ID 0xFF after 0x7E
     * so we will convert it here */
    if (id == 0xFF && last_id == 0x7E)
      id = 0x7F;

    /* match received packet with the table of sent packet info */
    matchPacket(id2seq(id), now_tv, rtp->RTP_header);

    last_id = id;
    break;
  case 1: /* listen mode */
    logDiffJitter(now_tv, rtp->RTP_header);
    break;
  default:
    return -1;
  }

  return 0;
} /* qosRTP() */

/* log statistic info from RR */
static int logRR(rtcp_report_block *rb, struct timeval *now_tv)
{
  static struct timeval rtt_tv;
  static uint32_t rtt_rtpt;
  unsigned long  rtt;

  /* ignore the first SR packet because dlsr is zero */
  if (rb->lsr) {
    rtt_rtpt = tv2rtptime(now_tv) - rb->lsr - rb->dlsr;
    rtt = (unsigned long)( ((rtt_rtpt & 0xffff0000) >> 16) * 1000000 
                           + (((rtt_rtpt & 0x0000FFFF) * 15625) >> 10) );
    rtptime2tv(&rtt_tv, rtt_rtpt);

    /* update rtt */
    if (g_stat.rtcp_rttMax < rtt)
      g_stat.rtcp_rttMax = rtt;
    g_stat.rtcp_rttAvr =
                      (g_stat.rtcp_rttAvr * (g_stat.recvRTCPPack - 1) + rtt)
                      / g_stat.recvRTCPPack;
  }

  fprintf(g_rtp.RTCPlog, "0x%08lX %u %u %lu %lu %ld.%05ld %lu %f %lu ",
          rb->ssrc,
          (unsigned)(((rb->highest_seqno) & 0xffff0000) >> 16),
          (unsigned)((rb->highest_seqno) & 0xffff),
          (lu)rb->lsr,
          (lu)rb->dlsr,
          rtt_tv.tv_sec, rtt_tv.tv_usec / 10,
          (lu)rb->jitter,
          rb->frac_lost / 256.0,
          (lu)rb->cum_lost);
  if (debug >=2 )
  fprintf(g_o.debfile,
"RR: ssrc=0x%08lX seq_cycles=%u seq=%u lsr=%08lX dlsr=%08lX last_rtcp=%ld.%06ld \
rtcp_count=%d rtt=%ld.%06ld rtt_hex=%08lX now_hex=%08lX jitter=%lu fract_lost=%f total_lost=%lu ",
          rb->ssrc,
          (unsigned)(((rb->highest_seqno) & 0xffff0000) >> 16),
          (unsigned)((rb->highest_seqno) & 0xffff),
          (lu)rb->lsr,
          (lu)rb->dlsr,
          g_stat.rtcp_sent_tv.tv_sec, g_stat.rtcp_sent_tv.tv_usec,
          g_stat.rtcp_sent_count,
          rtt_tv.tv_sec, rtt_tv.tv_usec,
          (lu)rtt_rtpt,
          (lu)tv2rtptime(now_tv),
          (lu)rb->jitter,
          rb->frac_lost / 256.0,
          (lu)rb->cum_lost);

  return 0;
} /* logRR() */

/* QOS measurement from RTCP packet
 * calculate QOS statistics from SR/RR */
static int qosRTCP(rtcp_packet *p, struct timeval *now_tv)
{
  int i, need_newline = 0;
  struct timeval ntp_tv;
  rtcp_sr *sr;

  switch (p->common->pt) {
  case RTCP_PACKET_SR:
    sr = p->variant.sr.sr_fixed;

    /* calc receive delay from SR *
     * arrival_time - SR_made_in_remote_ntp */
    ntp2tv(&ntp_tv, sr->ntp_secs, sr->ntp_frac);

    fprintf(g_rtp.RTCPlog, "%lu %lu %lu %ld.%06ld ",
                            sr->pkt_count,
                            sr->rtp_stamp,
                            sr->oct_count,
                            ntp_tv.tv_sec, ntp_tv.tv_usec);
    if (debug >=2 )
    fprintf(g_o.debfile,
"SR: pckts_sent=%lu rtp_ts=%lu bytes_sent=%lu ntp_hex=0x%08lX%08lX \
ntp_ts=%ld.%06ld now_ts=%ld.%06ld ",
                            sr->pkt_count,
                            sr->rtp_stamp,
                            sr->oct_count,
                            sr->ntp_secs, sr->ntp_frac,
                            ntp_tv.tv_sec, ntp_tv.tv_usec,
                            now_tv->tv_sec, now_tv->tv_usec);
    for (i = 0; i < RTCP_RC(*(p->common)); i++)
      logRR((rtcp_report_block *)p->variant.sr.rblock[i], now_tv);
    need_newline++;
    break;
  case RTCP_PACKET_RR:
    for (i = 0; i < RTCP_RC(*(p->common)); i++)
      logRR((rtcp_report_block *)p->variant.rr.rblock[i], now_tv);
    break;
    need_newline++;
  default:
    if (debug)
      fprintf(stderr, "compound RTCP packet with: %u\n", p->common->pt);
    break;
  } /* switch */

  if (need_newline) {
    putc('\n', g_rtp.RTCPlog);
    if (debug >=2 ) putc('\n', g_o.debfile);
  }

  return 0;
} /* qosRTCP() */

/***************************************************************************/
/** read RTP handler.
 * This is called by Notify function whenever receives a packet
 * on a RTP port */
static Notify_value readRTPHandler(Notify_client client, int dummyfd)
{
  struct timeval now_tv;
  char rBuff[BUFSIZE];
  int rBytes = BUFSIZE;
  rtp_packet rtp;
  rtperror err;

  /* record arrival time, RTP packet */
  gettimeofday(&now_tv, NULL);
  g_stat.rtp_lasttime = now_tv;

  /* read arrived RTP packet */
  err = RTPReceive(g_rtp.cid, g_rtp.rtp_fd, rBuff, &rBytes);
  if (g_o.verbose >= 2 && (err == RTP_OK || err == RTP_PACKET_LOOPBACK))
    printf("received RTP: port: %d counter: %d size: %d\n",
                                g_o.localPort, g_stat.recvRTPPack, rBytes);
  if (err != RTP_OK && (err == RTP_PACKET_LOOPBACK ? !g_o.loopback_ok : 1)) {
    if (debug)
      fprintf(stderr, "RTPReceive RTP (%d): %s\n",
                                      g_rtp.rtp_fd, RTPStrError(err));
    return NOTIFY_DONE;
  }

  rtp = RTPGetRTPPacket(rBuff, (long)rBytes);

  /* debug print, core dump the packet */
  if (debug >= 2)
    coreDumpPacket(rBuff, rBytes, 0);

  /* QOS measurement from RTP packet */
  if (qosRTP(&rtp, rBuff, rBytes, &now_tv) < 0) {
    fprintf(stderr, "%s: readRTPHandler(): qosRTP() error\n", g_o.argv0);
    return NOTIFY_DONE;
  }

  /* update RTP info */
  g_stat.prev_arrive_tv = now_tv;
  g_stat.prev_ts = rtp.RTP_header->ts;
  g_stat.recvRTPPack++;

  return NOTIFY_DONE;
} /* readRTPHandler() */

/** read RTCP handler.
 * This is called by Notify function whenever receives a packet
 * on a RTCP port */
static Notify_value readRTCPHandler(Notify_client client, int dummyfd)
{
  static int first_received;
  char rBuff[BUFSIZE];
  int rBytes = BUFSIZE;
  struct timeval now_tv;
  rtperror err;
  rtcp_packet rtcp;
  int i, rtcpnum;
  char *rtcps[MAX_RR_IN_COUMPOUND];

  /* record arrival time, RTCP packet */
  gettimeofday(&now_tv, NULL);
  g_stat.rtcp_lasttime = now_tv;

  /* read received RTCP packet */
  err = RTPReceive(g_rtp.cid, g_rtp.rtcp_fd, rBuff, &rBytes);
  if (g_o.verbose >= 2 && (err == RTP_OK || err == RTP_PACKET_LOOPBACK))
    printf("received RTCP: port: %d counter: %d size: %d\n",
           g_o.localPort + 1, g_stat.recvRTCPPack, rBytes);
  if (err != RTP_OK && (err == RTP_PACKET_LOOPBACK ? !g_o.loopback_ok : 1)) {
    fprintf(stderr, "RTPReceive RTCP (%d): %s\n",
                    g_rtp.rtcp_fd, RTPStrError(err));
    return NOTIFY_DONE;
  }
  /* validate RTCP packet */
  /* not implemented yet */

  /* debug print, core dump the packet */
  if (debug >= 2) {
    coreDumpPacket(rBuff, rBytes, 1);
  }

  rtcpnum = RTPSplitCompoundRTCP(rBuff, rtcps, rBytes);

  for (i = 0; i < rtcpnum; i++) {
    rtcp = RTPGetRTCPPacket(rtcps[i]);
    /* QOS measurement from RTCP packet */
    qosRTCP(&rtcp, &now_tv);
  }

  g_stat.recvRTCPPack++;

  /* store the time of the first RTCP packet */
  if (first_received == 0) {
    g_stat.rtcp_startime = now_tv;
    first_received++;
  }

  return NOTIFY_DONE;
} /* readRTCPHandler() */

/***************************************************************************/
/** send RTP packet. */
static int sendRtpPacket(void)
{
  int ts;
  QItems *qItem;
  rtperror err;
  static int marker = 1;    /* only the first packet sets the maker bit */
  char sBuff[BUFSIZE];

  ts = g_rtp.datasize;

  /* embed a noticeable ID in the payload */
  memset(sBuff, seq2id(g_rtp.rtp_seq), g_rtp.datasize);

  err = RTPSend(g_rtp.cid, ts, marker, g_o.payloadType,
                sBuff, g_rtp.datasize);
  if (err != RTP_OK && debug) {
    fprintf(stderr, "RTPSend(): %s\n", RTPStrError(err));
  }

  /* clear the marker bit */
  marker = 0;

  /* save seq and dep_tv in the queue table, only qos_mode == match */
  if (g_o.qos_mode == 0) {
    qItem = assembleItem(g_rtp.rtp_seq, &g_rtp.dept_tv);
    if (addQueue(g_stat.queue, qItem) == NULL) {
      fprintf(stderr, "%s: Waited too long to receive RTP packet at port: %d\n",
                      g_o.argv0, g_o.localPort);
      return 1;  /* terminate, queue table over flow */
    }
  }

  g_rtp.rtp_seq++;

  return 0;
} /* sendRtpPacket() */

/** send RTP handler.
 * This is called by Notify when timer expires for sending RTP packet */
static Notify_value sendRTPHandler(Notify_client client)
{

  /* send RTP packet */
  if (sendRtpPacket() == 1) {
    notify_stop();
    return NOTIFY_DONE;
  }

  /* set next departure time */
  timeradd(&g_rtp.dept_tv, &g_rtp.packInterval_tv, &g_rtp.dept_tv);

  /* check whether ending time has come */
  if (g_o.probe_sec != 0 && g_rtp.dept_tv.tv_sec > g_o.stop_tv_sec) {
    notify_stop();
    return NOTIFY_DONE;
  }

  /* next timer for sendHandler() */
  if (!flush)
    timer_set(&g_rtp.dept_tv, sendRTPHandler, (Notify_client)g_rtp.rtp_seq, 0);

  return NOTIFY_DONE;
} /* sendRTPHandler() */

/** send RTCP handler.
 * This is called by Notify when timer expires for sending RTCP packet */
static Notify_value sendRTCPHandler(Notify_client client)
{
  
  gettimeofday(&g_stat.rtcp_sent_tv, NULL);
  g_stat.rtcp_sent_count++;

  /* send RTCP packet */
  if (RTPExecute(g_rtp.cid, (rtp_opaque_t)client) < 0) {
    fprintf(stderr, "%s: sendHandler(): sendRtpPacket\n", g_o.argv0);
    return NOTIFY_DONE;
  }

  return NOTIFY_DONE;
} /* sendRTCPHandler() */

/** RTP schedule.
 * This is called by RTP library when needs to send RTCP packet 
 * just set the timer for sendRTCPHandler() and return */
void RTPSchedule(context cid, rtp_opaque_t opaque, struct timeval *tp)
{
  /* next timer for sendRTCPHandler() */
  if (!flush)
    timer_set(tp, sendRTCPHandler, (Notify_client)opaque, 0);

  return;
} /* RTPSchedule() */

/***************************************************************************/
/** flush queue.
 * stop receiving anymore packets and flushes the packets in the queue */
static void flushQueue(void)
{
  struct timeval timeout, wait;
  uint32_t i;

  /* very rough way but flushing queue buffer */
  if (timer_pending() != 0) {
    if (g_o.verbose) {
      puts   ("  Wait while flushing packets in the buffer.");
      putchar('\n');
    }
  }
  for (i = 0;; i++) {
    timeout.tv_sec  = wait.tv_sec  = 0;
    timeout.tv_usec = wait.tv_usec = 500000;
    timer_get(&timeout);
    if (timer_pending() == 0)
      break;

    /* sleep a little to avoid busy waiting */
    sleepMsec(500);

    if (debug)
      printf("flusshing: pending: %d\n", timer_pending());
  }
  return;
} /*** flushQueue() ***/

/** signal handler.
 * traps ctr-c and stop the emulator */
static Notify_value sigHandler(Notify_client client, int sig,
                               Notify_signal_mode mode)
{
  if (g_o.verbose == 0) goto NOPRINT; 
  putchar('\n');
  puts("!Terminated by ctr-c.");
  putchar('\n');
  puts  ("Closing network.");
  
  NOPRINT:
  flush++;
  flushQueue();
  notify_stop();

  return NOTIFY_DONE;
} /* sigHandler() */

/** stdin handler.
 * traps stdin and stop the emulator by the user's command */
Notify_value stdinHandler(Notify_client client, int dummyfd)
{
  char buff[STRSIZE];

  fgets(buff, STRSIZE, stdin);

  if (strcmp("bye\n", buff) == 0)
    goto TERMINATE;
  if (strcmp("quit\n", buff) == 0)
  if (strcmp("q\n", buff) == 0)
    goto TERMINATE;
  if (strcmp("stop\n", buff) == 0)
    goto TERMINATE;
  if (strcmp("exit\n", buff) == 0)
    goto TERMINATE;
  return NOTIFY_IGNORED;

  TERMINATE:
  printf("!Terminated by command: %s\n", buff);
  putchar('\n');
  puts  ("Closing network emulation.");

  flush++;
  notify_stop();
  flushQueue();

  return NOTIFY_DONE;
} /* stdinHandler() */

/***************************************************************************
 *** main ***
 ***************************************************************************/
/** main.
 * everything starts here */
int main(int argc, char *argv[])
{
  int i = 0;
  Notify_client client = 0;

  startupSocket();        /* for Win32 */

  /* Parse command line arguments */
  if (initOptions(&g_o, argv[0]) < 0) {
    fprintf(stderr, "%s: Initialize error.\n", argv[0]);
    return -1;
  }
  if (parseCmdLine(argc, argv) < 0) {
    return -1;
  }

  /** Now it is the real main routine **/
  /* Initialize */
  if (init(&g_o) < 0) {
    fprintf(stderr, "%s: Initialize error.\n", g_o.argv0);
    return -1;
  }

  /* Print title */
  if (printTitle(&g_o) < 0) {
    fprintf(stderr, "%s: Error: printing title.\n", g_o.argv0);
    return -1;
  }

  /* Set stdin handler (command from keyboard) */
  #ifndef WIN32
  if (!g_o.non_interactive)
    notify_set_input_func(client, stdinHandler, fileno(stdin));
  #endif

  /* Set signal handler (terminating)
   * win32 does not have SIGHUP but passes SIGINT instead */
  notify_set_signal_func(client, sigHandler, SIGINT, NOTIFY_ASYNC);
  notify_set_signal_func(client, sigHandler, SIGTERM, NOTIFY_ASYNC);
  #ifndef WIN32
  notify_set_signal_func(client, sigHandler, SIGHUP, NOTIFY_ASYNC);
  #endif

  /* Set readFd handler (calculate statistics of RTP) */
  if (notify_set_input_func((Notify_client)i, readRTPHandler, g_rtp.rtp_fd)
      != 0) { /* notify_set_input_func() does not reaturn meaningful value */
    fprintf(stderr, "%s: notify_set_input_func().\n", g_o.argv0);
    return -1;
  }
  /* Set readFd handler (calculate statistics of RTCP) */
  if (notify_set_input_func((Notify_client)i, readRTCPHandler, g_rtp.rtcp_fd)
      != 0) { /* notify_set_input_func() does not reaturn meaningful value */
    fprintf(stderr, "%s: notify_set_input_func().\n", g_o.argv0);
    return -1;
  }

  #ifndef DEBUG_NO_SEND
  /* set the first timer of sending RTP packet */
  timer_set(&g_rtp.dept_tv, sendRTPHandler, (Notify_client)g_rtp.rtp_seq, 0);
  #endif

  if ( (i = notify_start()) == -1) {    /* != NOTIFY_OK, bug notify_start()? */
    fprintf(stderr, "%s: Notifier error %d.\n", g_o.argv0, i);
    return -1;
  }

  printStatistics();

  if (destroy() < 0) {
    fprintf(stderr, "%s: Could not destroy all the handler.\n", g_o.argv0);
    return -1;
  }

  return 0;
} /* main() */

